START TRANSACTION;

-- Ce fichier contient la création des tables nécessaires pour
-- le fonctionnement de cette application.
USE bachelor;

-- Les modules sont contenus dans la table.
-- Un module correspond à une procédure de scrap
CREATE TABLE IF NOT EXISTS modules (
    id bigint AUTO_INCREMENT,       -- L'ID du module
    name text NOT NULL,             -- Le nom du module
    usesSelenium boolean NOT NULL,  -- Si le module utilise selenium ou pas
    startArguments json NOT NULL,   -- Les arguments attendus au démarrage de la tâche (dict name => type)

    PRIMARY KEY (id)
);

-- La queue continent les instances des modules.
-- Chaque instance reste jusqu'à son traitement complet,
-- c'est-à-dire le succès de la tâche ou N échecs.
CREATE TABLE IF NOT EXISTS tasks (
    id bigint AUTO_INCREMENT,       -- L'ID de la tâche
    module_id bigint NOT NULL,      -- Le module a essayer
    arguments json NOT NULL,        -- Les arguments de lancement
    tries int NOT NULL DEFAULT 0,   -- Essais
    next_task bigint,                    -- Tâche qui suit après celle-ci

    PRIMARY KEY (id),

    CONSTRAINT fk_tasks_module
		FOREIGN KEY (module_id)
		REFERENCES modules (id)
);

CREATE TABLE IF NOT EXISTS failed_tasks (
    id bigint AUTO_INCREMENT,
    module_id bigint NOT NULL,
    tries int NOT NULL DEFAULT 0,   -- Essais
    message text NOT NULL,

    PRIMARY KEY (id),

    CONSTRAINT fk_failedTasks_module
		FOREIGN KEY (module_id)
		REFERENCES modules (id)
);

CREATE TABLE IF NOT EXISTS task_performance (
    id bigint AUTO_INCREMENT,
    module_id bigint NOT NULL,
    time int NOT NULL,

    PRIMARY KEY (id),

    CONSTRAINT fk_tasksPerf_module
        FOREIGN KEY (module_id)
        REFERENCES modules (id)

);

COMMIT