import threading
import zipfile
from os import getenv as env

from flask import Flask, request, send_from_directory
from loguru import logger

from common.networking.radio import Radio
from common.utils.LoguruHandler import FlaskToLoguruHandler
from common.utils.io import getFilesInFileFolder

app = Flask(__name__)
bridge = Radio("127.0.0.1", int(env("HTTPHANDLER_EVENTS_PORT")), False)


@app.route("/", methods=["GET"])
# Shows all web API usages
def showUsage():
	routes = {
		"/pageUpdate": "Notifies the crawler that the webpage has been refreshed by the server.",
		"/pageChange": "Notifies the crawler that the webpage has changed its state.",
		"/shutdown": "Shuts down the server. Do not use until all jobs are finished.",
		"/": "Shows this web page."
	}

	output = "This is a management web API. Please do not use without proper reason to do so.<br>\n"
	for k, v in routes.items():
		output += k + " : " + v + "<br>\n"

	return output


@app.route("/pageUpdate")
def noticeOfPageUpdate():
	uuid = request.args["uuid"]
	__notify__(uuid, True, False)
	return "OK"


@app.route("/pageChange")
def noticeOfPageChange():
	uuid = request.args["uuid"]
	__notify__(uuid, False, True)
	return "OK"


@app.route("/shutdown")
def shutdown():
	func = request.environ.get('werkzeug.server.shutdown')
	func()
	return "OK"


# API - Used for sending data directly without HTML
@app.route('/files/all', methods=['GET', 'POST'])
def downloadAll():
	zf = zipfile.ZipFile(env("FILE_LOCKER") + 'output.zip', mode='w')

	files = getFilesInFileFolder()
	files.remove("output.zip")
	# noinspection PyBroadException
	try:
		for f in files:
			zf.write(f)
	except Exception as e:
		return "Failed ! " + str(e), 500
	finally:
		zf.close()

	return send_from_directory(directory=env("FILE_LOCKER"), filename="output.zip")


# API - Used for sending data directly without HTML
@app.route('/files/<filename>', methods=['GET', 'POST'])
def download(filename):
	return send_from_directory(directory=env("FILE_LOCKER"), filename=filename)


# Notifies the observers that a change happened
def __notify__(observerID, hasPageUpdated: bool, hasPageChanged: bool):
	logger.debug("Notifying...")
	bridge.talk(observerID, {"updated": hasPageUpdated, "changed": hasPageChanged})


# Runs the web server
def start(port):
	def actualStart():
		logger.info("Starting web handler on port " + str(port) + "...")
		app.logger.addHandler(FlaskToLoguruHandler)
		app.run(host="0.0.0.0", port=port)

	threading.Thread(target=actualStart).start()
