from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from loguru import logger


class Sleepers:
	def waitForXPathToAppear(self, driver, xPath, timeout=5):
		try:
			logger.info("Waiting for XPATH variable {}...".format(xPath))
			isReady = driver.find_element_by_xpath(xPath)
			WebDriverWait(driver, timeout).until(EC.visibility_of(isReady))
			WebDriverWait(driver, timeout).until(EC.invisibility_of_element(isReady))
		except StaleElementReferenceException:
			logger.error("Stale values, retrying.")
			self.waitForXPathToAppear(driver, xPath, timeout)

	def waitForURL(self, driver, url, timeout=10):
		if driver.current_url == url:
			return
		WebDriverWait(driver, timeout).until(EC.url_to_be(url))

	def waitForJSEvent(self, driver: WebDriver, script, args=None):
		if args is None:
			args = []
		return driver.execute_async_script(script, args)
