import os
import platform
import random
import traceback
import uuid
from enum import Enum
from os import getenv as env
from time import sleep

from loguru import logger
from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support.select import Select
from selenium.webdriver.support.wait import WebDriverWait

from client.spiders.seleniumspiders.Sleepers import Sleepers
from client.spiders.seleniumspiders.Utils import Utils
from common.networking.radio import Radio
from selenium.webdriver.support import expected_conditions as EC

RESET_PERIOD = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[1]/table/tbody/tr/td[2]/div/input'

FRENCH_PLEASE = '/html/body/table/tbody/tr[1]/td/div/form[1]/table/tbody/tr/td[2]/div/a[2]'

CLEAR_MERCHANDISE_SELECTION = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[5]/table/tbody/tr/td[2]/div/input'

DATE_SELECTION_YEAR_START = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/span[1]/div/table/tbody/tr[2]/td[2]/select'
DATE_SELECTION_YEAR_END = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/span[1]/div/table/tbody/tr[2]/td[4]/select'
DATE_SELECTION_MONTH_START = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/span[1]/div/table/tbody/tr[3]/td[2]/select'
DATE_SELECTION_MONTH_END = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/span[1]/div/table/tbody/tr[3]/td[4]/select'
DATE_SELECTION_SEND_MANIFESTO = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/span[1]/div/table/tbody/tr[4]/td[4]/div/a[1]'
HAS_CORRECTLY_SENT_MANIFESTO= '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[2]/table/tbody/tr/td/div/div[2]/table/tbody/tr'

MERCHANDISE_NUMBER_SELECTOR = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/div[1]/table[1]/tbody/tr[3]/td[2]/table/tbody/tr/td[3]/fieldset/table/tbody/tr/td[2]/label'
MERCHANDISE_SELECTOR = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/div[1]/table[1]/tbody/tr[3]/td[2]/table/tbody/tr/td[1]/div/input'
MERCHANDISE_SELECT_ALL = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/table[3]/tbody/tr/td[2]/a'
MERCHANDISE_TOTAL_COMMERCE_REMOVE_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[6]/table/tbody/tr/td/div/div[2]/table/tbody/tr/td[3]/input'
MERCHANDISE_SELECTED = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[6]/table/tbody/tr/td/div/div[2]/table/tbody/tr'

COUNTRY_SELECT_ALL = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/table[3]/tbody/tr/td[2]/a'
COUNTRY_TOTAL_COMMERCE_REMOVE_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/div[8]/table/tbody/tr/td/div/div[2]/table/tbody/tr/td[2]/input'

TOTAL_SELECTION_ID = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/table[1]/tbody/tr[3]/td/div/a[2]'

NEXT_STEP = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/table/tbody/tr[3]/td/div/a[1]'
NEXT_STEP_3 = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[2]/div/form/table/tbody/tr[3]/td/div/a[2]'

DONE_OPTIONS_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td/form/table[1]/tbody/tr/td[1]/a[1]'
DONE_TOGGLE_OPTIONS_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[1]/table/tbody/tr/td[2]/input'
DONE_ACTIONS_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/table/tbody/tr[1]/td/table/tbody/tr/td[1]/table/tbody/tr[2]/td[2]/div/div'
DONE_FILENAME = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[2]/table/tbody/tr[2]/td/div/div/table[1]/tbody/tr[3]/td[2]/input'
DONE_LOOK_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[2]/table/tbody/tr[1]/td/table/tbody/tr/td[3]/table/tbody/tr[2]/td[2]/div/div/span'
DONE_SHOW_EMPTY_CELLS_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[3]/table/tbody/tr[2]/td/div/div[5]/table[2]/tbody/tr/td/table/tbody/tr/td[2]'
DONE_EXPORT_BUTTON = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td[1]/form/div/div[2]/table/tbody/tr[2]/td/div/div/table[1]/tbody/tr[4]/td/a'
DONE_TABLE_ITEMS = '/html/body/table/tbody/tr[2]/td[1]/div/table/tbody/tr/td/form/div/div/div[1]/span[2]/table/tbody/tr'
DONE_URL = "https://www.gate.ezv.admin.ch/swissimpex/public/bereiche/waren/result.xhtml"


class Parser:
	# Class' internal machine state
	class SwissImpexState(Enum):
		DATE_SELECTION = 1  # Currently selecting a date
		DIRECTION_SELECTION = 2  # Currently selecting the merchandise direction (from the outside or from the inside)
		MERCHANDISE_SELECTION = 3  # Currently selecting the merchandise
		COUNTRIES_SELECTION = 4  # Currently selecting the countries (imp/exp)orting the selected merchandise
		TOTAL_SELECTION = 5  # Currently checking the previous selections
		RESULT = 6  # The obtained results

	# Class internal selection variables
	driver = None

	currentState: SwissImpexState = SwissImpexState.DATE_SELECTION
	selectedYear: int = None
	selectedMonth: int = None
	selectedMerchandise: str = None
	ownID: uuid = None

	# Connection to other processes and other machines
	radio: Radio = None
	ownPort: int = None

	def __init__(self, selectedMerchandise: str, selectedYear: int, selectedMonth: int):
		self.selectedMerchandise = str(selectedMerchandise)
		self.selectedYear = int(selectedYear)
		self.selectedMonth = Utils.monthNumberToMonthName(selectedMonth)
		self.ownID = str(uuid.uuid1())
		self.ownPort = random.randint(40000, 50000)
		self.radio = Radio("127.0.0.1", int(env("HTTPHANDLER_EVENTS_PORT")), True)

	# Adds a JS in the web page to make sure a few JS events are sent to the HTTP Handler.
	def registerJSEventHandler(self, port):
		s = Sleepers()
		script = """
			let callback = arguments[1];
			ice.onAfterUpdate(function(updates){
				let hasChangedOfPage = [...updates.firstElementChild.firstElementChild.children]
					.some(a => a.getAttribute("id") === "formSelectionMain");
				// Sending actual response to custom webserver
				var xmlHttp = new XMLHttpRequest();
				var uuid = '%s';
				let url = 'http://127.0.0.1:%d/' + (hasChangedOfPage ? 'pageChange' : 'pageUpdate');
				xmlHttp.open("GET", url + "?uuid=" + uuid, true);
                xmlHttp.send( "" );
			});
			callback(); // Remove Selenium's default handler
		""" % (self.ownID, port)
		s.waitForJSEvent(self.driver, script)

	# Waits until the webpage's ready to accept new commands
	def w(self):
		item = self.radio.listen()
		logger.debug("Received notification from web page")
		logger.debug("Page updated : {0}, page changed {1}".format(item["updated"], item["changed"]))

	# Clicks on the next step button
	def nextStep(self):
		logger.info("Next step...")
		xpath = NEXT_STEP if self.currentState == self.SwissImpexState.DATE_SELECTION else NEXT_STEP_3
		WebDriverWait(self.driver, 3).until(EC.visibility_of_element_located((By.XPATH, xpath)))
		nextStep = self.driver.find_element_by_xpath(xpath)
		nextStep.click()

	# Processes the date page
	def __processDate__(self):
		# Clearing all
		self.driver.find_element_by_xpath(RESET_PERIOD).click()
		self.w()

		# Selecting year
		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DATE_SELECTION_YEAR_START)))
		yStart = Select(self.driver.find_element_by_xpath(DATE_SELECTION_YEAR_START))
		yStart.select_by_visible_text(str(self.selectedYear))
		self.w()

		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DATE_SELECTION_YEAR_END)))
		yEnd = Select(self.driver.find_element_by_xpath(DATE_SELECTION_YEAR_END))
		yEnd.select_by_visible_text(str(self.selectedYear))
		self.w()

		# Selecting month
		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DATE_SELECTION_MONTH_START)))
		mStart = Select(self.driver.find_element_by_xpath(DATE_SELECTION_MONTH_START))
		mStart.select_by_visible_text(str(self.selectedMonth))
		self.w()

		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DATE_SELECTION_MONTH_END)))
		mEnd = Select(self.driver.find_element_by_xpath(DATE_SELECTION_MONTH_END))
		mEnd.select_by_visible_text(str(self.selectedMonth))
		logger.info(
			"Selected {0} {1}...Sending manifesto to server.".format(self.selectedMonth, self.selectedYear))
		self.w()

		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DATE_SELECTION_SEND_MANIFESTO)))
		sendManifestToServer = self.driver.find_element_by_xpath(DATE_SELECTION_SEND_MANIFESTO)
		sendManifestToServer.click()
		self.w()

		self.nextStep()
		self.w()
		self.currentState = self.SwissImpexState.DIRECTION_SELECTION

	# Processes the merchandise page
	def __processMerchandise__(self):
		# We first need to iterate until we find the corresponding merchandise
		logger.info("Setting up form...")
		WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, MERCHANDISE_NUMBER_SELECTOR)))
		self.driver.find_element_by_xpath(MERCHANDISE_NUMBER_SELECTOR).click()
		self.w()
		logger.info("Locating search bar...")
		searchBar = self.driver.find_element_by_xpath(MERCHANDISE_SELECTOR)
		logger.info("Located, sending search criteria...")
		searchBar.clear()
		searchBar.send_keys(self.selectedMerchandise)
		self.w()

		logger.info("Selecting resulting merchandise")
		merchandise = self.driver.find_element_by_xpath(MERCHANDISE_SELECT_ALL)
		merchandise.click()
		self.w()

		if len(self.driver.find_elements_by_xpath(MERCHANDISE_SELECTED)) < 1:
			logger.warning("Failed to select merchandise properly, retrying")
		elif len(self.driver.find_elements_by_xpath(MERCHANDISE_SELECTED)) > 1:
			element = self.driver.find_element_by_xpath(MERCHANDISE_SELECTED + "[1]/td[3]/input")
			element.click()

			# Forcing next step
			self.w()
			self.nextStep()
			self.w()
			self.currentState = self.SwissImpexState.COUNTRIES_SELECTION

	# Process the countries page
	def __processCountries__(self):
		logger.info("Selecting all countries")
		selectAll = self.driver.find_element_by_xpath(COUNTRY_SELECT_ALL)
		selectAll.click()
		self.w()

		if len(self.driver.find_elements_by_xpath(MERCHANDISE_SELECTED)) >= 2:
			logger.info("Removing wrong elements (total commerce)")
			self.driver.find_element_by_xpath(MERCHANDISE_TOTAL_COMMERCE_REMOVE_BUTTON).click()
			self.w()

		self.driver.find_element_by_xpath(COUNTRY_TOTAL_COMMERCE_REMOVE_BUTTON).click()

		logger.info("Ready to rock'n'roll")
		self.w()
		self.nextStep()

		# Done :)
		self.currentState = self.SwissImpexState.RESULT  # We skip total for now (not shown)
		logger.info("=======================================")
		logger.info("= SCRAPPING IS NOW OVER FOR THIS PART =")
		logger.info("=======================================")
		logger.info("Waiting for results...")

	# Downloads the document
	def parse(self, handlerPort: int):
		logger.info("Setting parsing environment...")
		logger.info("Setting browser...")
		exec_ext = "-mac" if platform.system() == "Darwin" else "-lin"
		self.driver = webdriver.Firefox(executable_path=os.getcwd() + "/resources/bin/geckodriver" + exec_ext,
		                                options=Utils.getBasicOptions())

		logger.info("Subscribing to HTTP Event handler...")
		self.radio.subscribe(self.ownID)

		logger.info("Navigating to Swiss Impex...")
		self.driver.get("https://www.gate.ezv.admin.ch/swissimpex/public/bereiche/waren/query.xhtml")

		logger.info("Forcing language to french...")
		for _ in range(5):
			if len(self.driver.find_elements_by_xpath(FRENCH_PLEASE)) > 0:
				self.driver.find_element_by_xpath(FRENCH_PLEASE).click()

		logger.info("Registering event emitter...")
		self.registerJSEventHandler(handlerPort)

		logger.success("Environment set, starting...")
		try:
			while self.currentState != self.SwissImpexState.RESULT:
				logger.info("Iterating over state...")
				if self.currentState == self.SwissImpexState.DATE_SELECTION:
					self.__processDate__()

				elif self.currentState == self.SwissImpexState.DIRECTION_SELECTION:
					# We don't do anything here. Should still work
					logger.info("Doing nothing to commerce trafic.")

					self.w()
					self.nextStep()
					self.w()
					self.currentState = self.SwissImpexState.MERCHANDISE_SELECTION

				elif self.currentState == self.SwissImpexState.MERCHANDISE_SELECTION:
					self.__processMerchandise__()
				elif self.currentState == self.SwissImpexState.COUNTRIES_SELECTION:
					self.__processCountries__()

			# Now we download the file
			if self.driver.current_url != DONE_URL:
				Sleepers().waitForURL(self.driver, DONE_URL, 30)
			logger.info("Preparing results...")

			while len(self.driver.find_elements_by_xpath(DONE_OPTIONS_BUTTON)) == 0:
				if len(self.driver.find_elements_by_xpath(DONE_TOGGLE_OPTIONS_BUTTON)) == 0:
					logger.warning("The page is not loaded properly yet ! Waiting for button.")
					WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DONE_OPTIONS_BUTTON)))
				else:
					logger.info("Apparently the export menu is already open, ignoring next statement")

			ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_OPTIONS_BUTTON)).pause(0.5)\
				.perform()

			logger.info("Setting export elements correctly...")
			WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DONE_LOOK_BUTTON)))

			ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_LOOK_BUTTON)).pause(0.5).perform()
			WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DONE_SHOW_EMPTY_CELLS_BUTTON)))
			ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_SHOW_EMPTY_CELLS_BUTTON)) \
				.pause(1.5).perform()

			logger.info("Ensuring result table is correct...")
			# We check the server received that particular order because sometimes it just ignores it
			for i in range(5):
				if len(self.driver.find_elements_by_xpath(DONE_TABLE_ITEMS)) <= 2:
					logger.warning("Result table does not show the empty cells, clicking again")
					ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_SHOW_EMPTY_CELLS_BUTTON)) \
						.pause(1.5).perform()

			logger.info("Getting ready to export...")
			WebDriverWait(self.driver, 5).until(EC.visibility_of_element_located((By.XPATH, DONE_ACTIONS_BUTTON)))
			ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_ACTIONS_BUTTON)).pause(0.5).perform()

			logger.info("Setting filename...")
			for i in range(5):
				if len(self.driver.find_elements_by_xpath(DONE_FILENAME)) > 0:
					break
				logger.warning("Stuborn web interface not responding, retrying")
				ActionChains(self.driver).click(self.driver.find_element_by_xpath(DONE_ACTIONS_BUTTON)).pause(
					0.5).perform()

			textbox = self.driver.find_element_by_xpath(DONE_FILENAME)
			textbox.clear()
			filename = "SwissImpex_{0}-{1}_{2}".format(
				self.selectedMonth, self.selectedYear, self.selectedMerchandise
			)
			textbox.send_keys(filename)

			for i in range(3):
				logger.info("Clicked download button, Downloading...")
				self.driver.find_element_by_xpath(DONE_EXPORT_BUTTON).click()

				sleep(5)  # This allows the download to start
				if os.path.isfile(env("FILE_LOCKER") + filename + ".xlsx"):
					logger.success("All good.")
					return {"swissImpexOutputfileName": env("FILE_LOCKER") + filename + ".xlsx"}
				else:
					logger.error("File {} didn't download, retrying".format(filename))

			raise RuntimeError("Didnt manage to download {} properly".format(filename))

		except Exception as e:
			logger.error("[CRASH] Crashed !")
			name = "oops-{0}.png".format(random.randint(0, 10000000))
			self.driver.save_screenshot(env("FILE_LOCKER") + name)
			logger.error(traceback.format_exc())
			logger.error("Saved file state in " + name + ".")
			raise AssertionError(str(e))
		finally:
			self.driver.quit()
			logger.info("Done. Good night !")
