import time
from threading import Thread

from loguru import logger
from sqlalchemy import func
from sqlalchemy.orm import Session

from common.mvc import getSession
from os import getenv as env

# Allows for client performance observation
from common.mvc.Models import Task
from common.networking.intercom import Intercom


class PerformanceObserver:
	shouldStop: bool = False  # Should stop monitoring
	clients: list = []  # Clients list
	timing: int = 1  # Timing between every measure

	def __init__(self, clients, timing):
		self.folder = env("FILE_LOCKER")
		self.clients = clients
		self.timing = timing

	def start(self, shouldMonitorTasksInDB, shouldMonitorFilesInFileLocker):
		Thread(target=self.monitor, args=(shouldMonitorTasksInDB, shouldMonitorFilesInFileLocker)) \
			.start()

	# Monitors every N second the requested data
	# Format : number of seconds passed,
	def monitor(self, shouldMonitorTasksInDB, shouldMonitorFilesInFileLocker):
		if not shouldMonitorTasksInDB and not shouldMonitorFilesInFileLocker:
			logger.error("You have chosen not to monitor anything...")
			exit(250)

		logger.info("Starting performance observer : DB monitoring = {}, File monitoring = {}"
		            .format(shouldMonitorTasksInDB, shouldMonitorFilesInFileLocker))

		secondsPassed = 0
		lastFileCount = 0
		with open('measures.csv', 'a') as fd:
			while not self.shouldStop:
				currentRun = [str(secondsPassed)]

				if shouldMonitorTasksInDB:
					db = getSession()
					currentRun.append(str(db.query(func.count(Task.id)).scalar()))
					db.close()

				if shouldMonitorFilesInFileLocker:
					total = 0
					# Reaching all clients to get statrep files number
					for client in self.clients:
						intercom = Intercom(client, env("CLIENT_PORT"), True)
						intercom.talk({"type": "statrep"})
						statrep = {}
						try:
							statrep = intercom.listen(750)
						except:
							total = lastFileCount  # Safe
							break
						total += len(statrep.get("files") or [])
						lastFileCount = total
					currentRun.append(str(total))

				fd.write(",".join(currentRun) + "\n")
				fd.flush()
				time.sleep(self.timing)
				secondsPassed += 1

	def stop(self):
		self.shouldStop = True
		logger.info("Stopped observer with success.")
