import os.path
import re
from collections import OrderedDict
from os import path

import numpy as np
from loguru import logger
import pandas as p
from pandas import DataFrame


class XLSXParser:
	file: str = None
	data = None
	workbench: DataFrame = None

	def __init__(self, file):
		self.file = file
		if not path.exists(file) or not os.path.isfile(file):
			raise RuntimeError("No such file {}".format(path))
		self.data = {}

	def run(self, instructions: OrderedDict, index=None, assistedSpliceData=None) -> dict:
		regex = "([A-Z]+)([0-9]+)"
		alpha0 = numeric0 = alpha1 = numeric1 = None
		if assistedSpliceData is not None:
			alpha0, numeric0 = re.findall(regex, assistedSpliceData[0])[0]
			alpha1, numeric1 = re.findall(regex, assistedSpliceData[1])[0]
		self.workbench = p.read_excel(self.file, sheet_name=index if index is not None else 0,
		                              usecols="{}:{}".format(alpha0, alpha1) if assistedSpliceData is not None else None)

		logger.debug("Removing NaN values...")
		self.workbench = self.workbench.fillna("")

		if numeric0 is not None and numeric1 is not None:
			logger.info("Splicing data vertically from {} to {}".format(numeric0, numeric1))
			self.data = self.workbench.values[int(numeric0):int(numeric1)]

		for instruction in instructions:
			action = list(instruction.keys())[0]
			arguments = list(instruction.values())[0]
			if isinstance(arguments, list):
				if action == "take" and len(arguments) == 2:
					if not arguments[0].isnumeric() or arguments[1].isnumeric():
						logger.error("Take instructions are not numeric. Take is only for vertical data selection.")
					else:
						logger.info("Taking from {} to {} ...".format(arguments[0], arguments[1]))
						self.data = self.data[arguments[0]:arguments[1]]

				elif action == "replace" and len(arguments) == 2:
					logger.info("Replacing from '{}' to '{}'...".format(arguments[0], arguments[1]))
					self.data = np.where(self.data == arguments[0], arguments[1], self.data)

			elif isinstance(arguments, str):
				if action == "export" and isinstance(arguments, str):
					dest = os.path.splitext(self.file)[0] + ".csv"
					# noinspection PyTypeChecker
					np.savetxt(dest, self.data, delimiter=';', fmt='%s')
					self.data = {"dest": dest}

			elif isinstance(arguments, bool):
				if action == "cleanup" and type(arguments) == bool and arguments:
					os.remove(self.file)

			else:
				logger.error("Incompatible instruction type {} {}".format(type(action), arguments))

			logger.success("Done.")

		logger.success("Finished instructions for this file.")
		return self.data

