# Méthodologie

Ce chapitre décrit la conception du framework Inari. Nous commencerons par parler en détail de l'architecture, en commençant par l'infrastructure du projet, son langage de programmation, des structures observables; nous continuerons en évoquant les structures de données implémentées, pour enfin discuter de la communication du projet.
Une fois la partie architecture réalisée, nous décrirons la réalisation de ce projet, en donnant deux exemples de systèmes de scraping implémentés dans ce projet.

## Architecture du projet

L'infrastructure de ce projet a été pensée pour être flexible. En effet, la récupération de données par scraping peut être très longue.
En proposant une architecture flexible, les pannes sont tolérables et n'impacteront qu'une fraction du projet. Par exemple, si on possède plusieurs machines dédiées au téléchargement et que l'une d'entre elles devait cesser de fonctionner, les autres pourront continuer à travailler.
De cette façon, il est tout de même possible de récupérer dans la majorité des cas une partie des informations scrapées. Cela est particulièrement utile sur les scraping de très longue durée.
En plus d'être flexible, le but final d'Inari est d'éviter au maximum à l'utilisateur de devoir programmer. Le système proposera des _modules_, capables chacun d'effectuer une activité précise; l'utilisateur pourra ensuite les lier pour créer son programme à la carte. Il pourra ainsi par exemple indiquer de télécharger des données spécifiques du site Swiss-Impex, site sur lequel nous reviendrons. Ces données sont téléchargées au format XLSX (Excel); on peut ensuite les convertir ensuite dans un format facilement lisible comme le format (+CSV_a) pour enfin supprimer les anciens fichiers Excel.

Les modules ne sont, du point de vue de l'utilisateur, que des boîtes noires [^10]. L'utilisateur entre un certain nombre de paramètres, et un fichier sort du module. Il n'a ainsi pas besoin de se préoccuper de la partie programmation.

Du côté utilisateur, on n'attend pas de l'utilisateur qu'il sache programmer. Un fichier, nommé _playbook_, décrit chaque action à effectuer et dans quel ordre. Il suffit à l'utilisateur d'utiliser le manuel d'utilisation d'Inari pour savoir s'en servir. L'annexe un donne un exemple de _playbook_ sur lequel [nous reviendrons](#Application-pratique-complete).

[^10]: En vérité, il ne s'agit pas véritablement de boîtes noires comme l'entendrait par exemple Stallman [@williams_hackers_2002]. L'utilisateur serait tout à fait en mesure d'aller voir comment fonctionne le module et le modifier s'il le souhaitait.

### Infrastructure

Scraper un site est une action qui peut être longue. Sur des sites complexes embarquant du JavaScript pour naviguer sur le site, il est nécessaire d'attendre entre chaque action un évènement JavaScript confirmant que l'action a bien été prise en compte. En comparaison à un site simple, cela peut prendre beaucoup plus de temps. De plus, tout processus long a un risque plus élevé de s'arrêter de fonctionner
qu'un processus plus court. Les scrapers sont déjà connus pour être fragiles et planter si une mise à jour du site est effectuée. Si la structure du site est touchée (et elle l'est automatiquement lors d'une mise à jour d'un site), les chances de plantage d'un scraper quel qu'il soit seront hautes.
Il est intéressant pour d'énormes masses de données de répartir la charge de travail. Ce travail propose deux répartisseurs: le premier, nommé _Schnell_[^11], permet de créer plusieurs processus (qu'on appelle _workers_, ou _travailleurs_) sur une même machine. Le second, quant à lui, permet de répartir la quantité de travail sur plusieurs machines. Ces deux répartitions permettent d'effectuer un grand nombre de tâches en parallèle.
Les machines travaillant sur une ou plusieurs tâches sont considérées comme les machines _clientes_ : elles n'ont aucun contrôle sur le flux d'exécution et ne connaissent jamais les autres machines.
Une autre machine, quant à elle, est considérée comme le _serveur_ : c'est elle qui va envoyer les tâches aux clients, qui va décider si un client doit s'arrêter et qui peut récupérer les documents téléchargés de chaque client. Le serveur respecte la tolérance aux pannes de l'infrastructure; en cas de plantage, seul le serveur cessera de fonctionner car une fois les tâches envoyées, les clients ne dépendent plus du serveur.
Enfin, un second serveur est nécessaire au fonctionnement optimal du service : une base de données permettant de stocker les tâches. Même si la base de données plante, les tâches, effectuées ou non, seront conservées.

\cimg{figs/infra.png}{scale=0.45}{Infrastructure du projet}{Source : réalisé par Théo Pirkl}

Le serveur est assez simpliste. Une fois démarré, il va lire un fichier (que l'on nomme un playbook) qui permettra de créer les tâches à envoyer aux clients, qu'il distribuera de façon équitable. Une fois ces dernières distribuées, il reste allumé dans le seul but de laisser un accès au panneau de contrôle. Ce dernier permet de récupérer les fichiers téléchargés par les clients.

Les clients, en revanche, sont plus complexes. Ils sont à l'écoute en permanence de nouvelles commandes de la part du serveur, mais doivent en plus supporter toute l'infrastructure pour les différents processus (en l'occurence, ceux téléchargeant les documents).
Chaque client exécute $N$ processus, où $N$ est fixé à l'avance. Ces $N$ processus sont ici pour effectuer les tâches envoyées par le serveur de façon concurrente (c'est l'implémentation du multiprocessing). 
Un autre processus, dit "principal", contient deux threads : un en charge de la communication vers l'extérieur et un autre en charge de gérer les communications entre les processus travaillant sur une tâche. Cette configuration, bien que compliquée, permet de gérer n'importe quelle tâche de façon aisée. La figure 2.2 illustre l'architecture client.

\cimg{figs/client-infra.png}{scale=.3}{Infrastructure interne du client}{Source : réalisé par Théo Pirkl}

[^11]: _Schnell_ signifie _vite_ ou _rapidement_ en allemand.

### Langage de programmation choisi

Le langage choisi pour ce projet est Python. Python est un langage particulièrement populaire pour deux raisons : sa facilité de lecture et sa capacité à effectuer des tâches complexes (ce malgré sa simplicité).  
Python est considéré comme le second langage recommandé à apprendre après le HTML [@alison_10_nodate]; c'est donc le premier langage de programmation polyvalent recommandé aux débutants. Il est tellement populaire qu'il est choisi quasiment systématiquement dans les projets de recherche dans l'(+IA_a).
Un de nos buts étant de proposer quelque chose d'accessible au public, il semble logique de choisir un langage qui l'est aussi.

En outre, des centaines de milliers de paquets et de librairies Python existent déjà, rendant la tâche d'interconnecter Inari à un système extrêmement facile dans bien des cas.
Au vu de ces faits, Python est un candidat idéal à l'implémentation de ce projet.

### Technologies choisies

Nous avons parlé dans le chapitre précédent des technologies en commun entre les (+SAAS_a) et les framework. Certaines technologies existantes ont été réutilisées dans ce projet. La technologie XPath est un excellent exemple : tous les modules utiliseront pour se repérer des expressions XPath. Il n'y a pas plus efficace et les XPath sont relativement bien documentées, raisons pour lesquelles les XPath seront utilisées dans ce projet.
Nous nous servirons aussi des technologies de multiprocessing, car elles permettent d'accélérer de longues tâches : dans le cadre du scraping, chaque seconde de gagnée est précieuse.
Enfin, nous nous servirons aussi comme tous les autres logiciels de scraping d'un navigateur : nous utiliserons Firefox, installé sur les machines de l'HEPIA et suffisamment répandu sur les distributions Linux pour ne pas poser de problème de compatibilité. Selenium WebDriver sera le contrôleur permettant au code Python de gérer Firefox. Selenium WebDriver est un contrôleur régulièrement utilisé pour faire du scraping, raison pour laquelle il a été choisi.

### Modèles de données

Différents modèles sont implémentés dans ce projet. Les deux principaux sont les tâches et les modules : un module est en charge d'une action spécifique (les boîtes noires que nous avons cité plus haut). Une tâche, quant à elle, est une instance d'un module. Un module est défini par un numéro d'identification unique, un nom et un fanion déterminant si le module doit effectuer une tâche complexe ou non et un tableau contenant les arguments attendus pour exécuter le module en question.

Les tâches sont définies elles aussi par un numéro d'identification unique, par le numéro unique du module que représente la tâche, les arguments à fournir au module pour fonctionner et le nombre de fois que cette tâche a été effectuée. Chaque tâche est effectuée au moins une fois. Si la tâche est accomplie avec succès à la première exécution, alors la tâche est supprimée, sinon elle est relancée. Elle sera relancée au maximum trois fois, puis mise dans la table des tâches en échec. De cette façon il est possible de déterminer quelle tâche a échoué systématiquement et pourquoi (le dernier message d'erreur est inclus avec la tâche).

\cimg{figs/inari-error.png}{scale=.4}{Exemple de gestion d'erreur}{Source : réalisé par Théo Pirkl}

Enfin, Inari effectue des mesures de performance pour chaque tâche, ce afin de déterminer l'efficacité du système dans son ensemble. Les mesures observées seront affichées dans la section résultats de ce rapport.

### Communication

La communication est faite de deux façons : dans la très grande majorité des cas, on utilise ZeroMQ[^12] (ou ZMQ), qui permet une communication inter-machine
et inter-processus. ZeroMQ est particulièrement populaire, propose une implémentation en Python et est extrêmement bien documenté. Il possède différents modèles de communication dans lesquels nous ne rentrerons pas en détail; ces derniers permettent en somme d'émettre un message à un ou plusieurs destinataires.

[^12]: Un librairie d'envoi et de réception de messages open-source.

Quand la communication avec ZMQ n'est pas souhaitable ou possible (cas sur lequel nous reviendrons), le protocole (+HTTP_a) est utilisé : le panneau de contrôle utilise par exemple exclusivement HTTP.
Les scrapers utilisent aussi (+HTTP_a) car il ne leur est pas possible d'utiliser ZMQ. Pour gérer certains évènements générés par les pages web qu'ils visitent, il est nécessaire d'injecter du JavaScript dans la page web en question. Ainsi, si la page web fonctionne en partie ou totalement avec JavaScript,
il est possible d'intercepter les évènements signalant qu'une action s'est produite et agir en conséquence.
Utiliser ZMQ dans cette situation aurait été bien plus compliqué qu'utiliser (+HTTP_a). Qui plus est, JavaScript permet d'envoyer des appels HTTP. Il suffit donc de créer un petit serveur web capable de réceptionner les appels HTTP lancés par JavaScript. Ainsi, on est tout à fait capable de récupérer les évènements générés par la page web dans notre infrastructure. Un seul serveur web de "surveillance" est démarré par client afin d'éviter de surcharger la machine.

\cimg{figs/js-infra.png}{scale=.4}{Gestion des évènements JavaScript}{Source : réalisé par Théo Pirkl}

L'avantage de gérer JavaScript est énorme. On estime en 2020 que 95% des sites web se servent de ce langage [@noauthor_usage_nodate] . Dans certains cas, il n'est utilisé qu'à des fins esthétiques (ouverture de menus, animations, etc).
Dans d'autres, toute la navigation du site en dépend. Dans tous les cas, la capacité à pouvoir gérer ce langage est cruciale pour un scraper se voulant modulaire.

### Génération des tâches

Chaque tâche est générée selon le playbook fourni par l'utilisateur. Le playbook consiste en une liste (+YAML_a) de modules. Chaque module spécifié dans le playbook est analysé et génère un nombre de tâches soit pour le client, soit pour le serveur. C'est le serveur qui génère les tâches et les distribue aux clients. Chaque module spécifié dans le playbook peut inclure un nombre variable de paramètres simples ou complexes. Les paramètres complexes permettent de rapidement générer une série de valeurs. Prenons un exemple :

\cimg{figs/playbook-complex.png}{}{Exemple de paramètre complexe}{Source : réalisé par Théo Pirkl}

Le module Swiss-Impex va ici considérer le paramètre _year_ et voir qu'une valeur _start_ et _end_ ont été définies. Il va alors automatiquement remplacer _year_ par toutes les valeurs possibles entre 1989 et 2020 (2020 non inclus). En voyant ensuite _month_, il fera de même. Une fois cela fait pour les tous les paramètres ayant défini _start_ et _end_, le système générera toutes les combinaisons de valeurs possibles : Swiss-Impex avec _month_ à 0, _year_ à 1989, Swiss-Impex avec _month_ à 1, _year_ à 1989, etc. La syntaxe _start_ et _end_ est une syntaxe propre au projet (aux meilleures connaissances de l'auteur), mais le langage utilisé et attendu pour le playbook est le (+YAML_a). 

Cela permet de générer un très grand nombre de tâches en très peu de lignes.

## Réalisation

Différents sites web ont été analysés et traités par Inari. Ces sites ont été choisis sur la base de besoins concrets.

| Nom du site                                                            | Description                                                                                                                                                                                                                                                                                                                                                                         |   |   |
|------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|---|---|
|                FAO Genève                |                      La Feuille d'Avis Officielle (ou FAO) publie tous les jours des informations  sur ces citoyens et entreprises en tout genre. Il peut être intéressant de s'y abonner si l'on souhaite recevoir des nouvelles sur certains  mots-clés. Le site ne propose toutefois aucun système d'abonnement et n'en semble pas disposé à en ajouter un.                     |   |   |
| Swiss-Impex | Swiss-Impex propose toutes les statistiques d'import/export de la Suisse. Les informations sur ce site  ont déjà permis de détecter des irrégularités et de dévoiler des scandales. Les informations de ce site  sont difficiles à récupérer automatiquement, et elles sont en très grand nombre.  L'automatisation de la récupération d'informations de ce site serait intéressant. |   |   |
|                                                                        |                                                                                                                                                                                                                                                                                                                                                                                     |   |   |
Table: Etude de deux sites intéressants à traiter par Inari.

\begin{center}
    \textit{Source : réalisé par Théo Pirkl}
\end{center}

Ces deux sites proposent des données étatiques permettant de donner des informations tout à fait intéressantes : des sites en font même la collecte pour les revendre[^13].

[^13]: [Lex Oculus](https://www.lexoculus.ch), [Easy Monitoring](https://www.easymonitoring.ch/fr/surveillance-fosc)

## Catégorie un : Scrapers simples

On définit par un _scraper simple_ un scraper n'ayant pas besoin d'intervenir dans le JavaScript pour traiter un site. Concrètement, cela signifie qu'il suffit de cliquer sur des éléments hypertexte (boutons, images, liens) pour arriver au-x document-s souhaité-s.
La catégorie un regroupe donc tout site ne dépendant pas de JavaScript pour naviguer à travers les pages.
Inari utilise un framework existant, nommé Scrapy, capable de traiter les sites web rentrant dans la catégorie un.
Ce dernier permet le téléchargement très rapide[^14] de documents sur un site web.

[^14]: Pour plus d'informations sur les performances de Scrapy, voir la section mesures

### Application pratique : FAO Genève

La FAO Genève est un excellent exemple de scraper simple. JavaScript n'est utilisé qu'à des fins esthétiques. On peut donc utiliser Scrapy pour ce genre de site.
L'implémentation avec Scrapy est relativement simple : en une cinquantaine de lignes, on est capable de récupérer les documents du site.
Scrapy étant lui-même son framework, aucun module n'est nécessaire. On peut démarrer avec une configuration réduite un scraper rentrant dans la catégorie un, c'est-à-dire démarrer avec un seul client, au vu de la vitesse de traitement.
Les modules présentés ci-dessus sont tout à fait compatibles avec Scrapy, mais ils ne sont pas utilisés au vu de la seule efficacité de ce framework.

Dans le cas de la FAO Genève, on ne trie pas les documents, mais on les récupère en bloc. Il est toutefois tout à fait possible de récupérer les documents en les filtrant avant de les récupérer.

## Catégorie deux : Scrapers complexes

On définit par _scraper complexe_ un scraper ayant obligatoirement besoin de JavaScript pour traiter et naviguer sur un site. Dans cette configuration, tous les modules mentionnés plus tôt sont nécessaires au traitement de documents : la complexité étant plus importante, chaque document prend plus de temps à être récupéré.
Aucun framework de scraping n'est ici utilisé. Pour faire fonctionner un scraper de catégorie deux, on démarre un navigateur (dans notre cas, Firefox) qui va, comme son nom l'indique, naviguer sur le site qu'on veut traiter. On donne au navigateur les instructions à effectuer et elles sont lancées séquentiellement.

### Application pratique : Swiss-Impex

Swiss-Impex fonctionne sur une technologie Java-JavaScript nommée IceFaces. IceFaces se base sur des appels (+AJAX_a) qui oblige le support de JavaScript.
Par conséquent, la catégorie deux de scrapers est idéale. Le module _SwissImpex Parser_ interagit avec le navigateur et est capable de se diriger jusqu'au document recherché. Ici, chaque action est envoyée au serveur Swiss-Impex. Il est donc nécessaire d'attendre que chaque action soit acquittée par le serveur avant d'effectuer le prochain clic. Le module de communication JavaScript est donc ici utilisé pour garantir une réception des évènements JavaScript : ici, on s'intéresse aux acquittements du serveur Swiss-Impex.

L'annexe un donne un exemple de _playbook_ dans lequel on peut voir Swiss-Impex apparaître. Ce playbook propose d'effectuer les tâches suivantes :

- Lire le fichier contenu dans `./resources/EUMerchIDs.txt` (ce fichier contient l'identificateur de chaque catégorie de marchandises catégorisé en Europe)
- Filtrer les identificateurs et garder ceux respectant une expression régulière[^15]
- Envoyer aux clients (dans le cas de l'annexe un, il y en a deux) un ensemble de tâches Swiss-Impex également réparti.
- Lors de la réception des tâches Swiss-Impex, le client va les envoyer dans l'ordre de réception (ce qui va aboutir au téléchargement du document au format Excel)
- A la fin de chaque tâche Swiss-Impex, le fichier XLSX est converti en CSV pour faciliter la lecture; Enfin, le fichier Excel est supprimé.

Chaque instruction "attend" la précédente : le travail est donc fait de façon séquentielle. Toutefois, chaque travailleur n'en attend aucun autre : autrement dit, chaque processus va séquentiellement effectuer les tâches spécifiées dans le _playbook_ de son côté.

\cimg{figs/inari-swissimpex.png}{scale=0.3}{Représentation graphique d'un lancement Swiss-Impex}{Source : réalisé par Théo Pirkl}

Dans le cas de Swiss-Impex, on choisit les documents qu'on souhaite recevoir : on choisira dans les mesures de récupérer tous les documents liés à une catégorie de marchandises, puis de récupérer tous les documents de toutes les catégories.

[^15]: Une expression régulière, ou Regex (ou encore RegExp) est une chaîne de caractères décrivant un ensemble de chaînes possibles.

## Bilan

L'infrastructure d'Inari n'est pas évidente. Elle possède plusieurs modules, relativement lourds pour le client et n'est pas aisée à comprendre.
Toutefois, du point de vue de l'utilisateur, c'est une solution facile (du moment que le module existe dans Inari), car il n'a pas à programmer.
Le support de JavaScript est aussi un énorme plus car il permet de supprimer les barrières qu'on pourrait observer en tentant de scraper un site : on peut donc théoriquement grâce à Inari scraper n'importe quel site.

\pagebreak